/* 
 *          This source file is part of DOSRUN. GNU/GPL source code.
 *                 (C)1999 LGB, Gabor Lenart, lgb@vlug.vein.hu
 *                         A bit confusing in its name,
 *   this is generally for Linux<->DOS name translating, but there're much
 *                  about DOS drive emulation inside as well.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include "dosrun.h"
#include "config_dosrun.h"
#include "filename.h"
#include "derrno.h"


char path[DOS_PATH_MAXSIZE];

static int current_drive;
static int dosdrives;
int last_drive; /* this carries the last drive id was used at path_d2u function */

static struct drive_struct drives[CONFIG_MAX_DRIVES];


int dir_isdosroot ( int drive , char *dir )
{
  int size=strlen(dir);
  if (size>1&&dir[size-1]=='/') size--;
  return !strncmp(dir,drives[drive].root,size);
}


/* Converts DOS path into UNIX, check [dosrun acl] access right */
/* TODO: process device DOS files : con,nul,aux,prn ... */
int path_d2u ( char *s , int access )
{
  int max=DOS_PATH_MAXSIZE;
  char wpath[DOS_PATH_MAXSIZE];
  char *t;
  if (s[1]==':') {
    if (*s>='a'&&*s<='z') last_drive=*s-'a';
      else if (*s>='A'&&*s<='Z') last_drive=*s-'A';
        else return DNODRIVE; /* illegal drive letter */
    s+=2;
    max-=2;
  } else last_drive=current_drive;
  if (last_drive>=CONFIG_MAX_DRIVES||!drives[last_drive].valid) {
    if (stddebug) fprintf(stddebug,"FILENAME: %c: illegal drive.\n",last_drive+'A');
    return DNODRIVE; /* illegal drive */
  }
  /* some UGLY ACL checking follows */
  if ((access&WRITE_ACCESS &&!drives[last_drive].acl.write )||
      (access&DELETE_ACCESS&&!drives[last_drive].acl.delete)||
      (access&MOVE_ACCESS  &&!drives[last_drive].acl.move  )||
      (access&CHMOD_ACCESS &&!drives[last_drive].acl.chmod )||
      (access&CREATE_ACCESS&&!drives[last_drive].acl.create)) {
    if (stddebug) fprintf(stddebug,"FILENAME: ACCESS DENIED debug=%s drive=%d (root=%s)\n",s,last_drive,drives[last_drive].root);
    return DREADONLY;
  }
  if (stddebug) fprintf(stddebug,"FILENAME: debug=%s drive %c: (root=%s)\n",s,last_drive+'A',drives[last_drive].root);
  strcpy(wpath,drives[last_drive].root);
  if (*s!='\\') strcat(wpath,drives[last_drive].dir);
  t=strlen(wpath)+wpath;
  *(t++)='/';
  while (*s&&max--)
    if (*s=='\\') *(t++)='/',s++; else *(t++)=tolower(*(s++));
  *t='\0';
  realpath(wpath,path);
  /* now check for drive boundary */
  if (strncmp(path,drives[last_drive].root,strlen(drives[last_drive].root))) {
    if (stddebug) fprintf(stddebug,"FILENAME: file outside drive %c: wpath=%s path=%s\n",last_drive+'A',wpath,path);
    return DNOPATH;
  }
  if (stddebug) fprintf(stddebug,"FILENAME: file OK %c: wpath=%s path=%s\n",last_drive+'A',wpath,path);
  return 0;
}


/* broken function to get number of floppy drives.
   we simply assmues A: and B: to be fdds, so if they're installed
   by dosrun configuration the number is counting both of them */
int dosdrive_getnumoffdds ( void )
{
  return drives[0].valid+drives[1].valid;
}


/* DOS style function. drive=0 (actual) 1=A: 2=B:
   respond at 'path' in DOS format suitable for get_cwd dos function */
int get_dosdir ( int drive , char *p )
{
  char *s;
     char *backup=p;
  if (drive) drive--; else drive=current_drive;
  if (drive>=CONFIG_MAX_DRIVES||!drives[drive].valid) return DNODRIVE;
  s=drives[drive].dir;
  if (*s=='/') s++;
  while (*s)
    if (*s=='/') *(p++)='\\',s++; else *(p++)=toupper(*(s++));
  if (*(p-1)=='/') *(p-1)='\0'; else *p='\0';
  if (stddebug) fprintf(stddebug,"DOSDRIVE: get cwd on %c: = \"%s\"\n",drive+'A',backup);
  return 0;
}


/* there is NO drive letter checks, assuming it's called from trusted function ... */
void set_dosdir_unix ( int drive , char *s )
{
  drives[drive].dir=dosrun_realloc(drives[drive].dir,strlen(s)+1);
  strcpy(drives[drive].dir,s);
  if (stddebug) fprintf(stddebug,"DOSDRIVE: (unix): drive %c: cwd changed to %s.\n",drive+'A',s);
}



int set_dosdir ( char *s )
{
  int r=path_d2u(s,DIR_ACCESS);
  if (r) return r;
  if (access(path,X_OK|R_OK)) return DNOACCESS;
  set_dosdir_unix(last_drive,path+strlen(drives[last_drive].root));
  if (stddebug) fprintf(stddebug,"DOSDRIVE: setting cwd for %c: (root=%s) to \"%s\" \"%s\"\n",last_drive+'A',drives[last_drive].root,path,path+strlen(drives[last_drive].root));
  return 0;
}


__inline__ void dosdrive_init ( void )
{
  char *home=getenv("HOME");
  bzero(drives,sizeof(drives));
  dosdrives=0;
  dosdrive_add(2,"/");
  getcwd(path,DOS_PATH_MAXSIZE);
  set_dosdir_unix(2,path);  /* only valid here because root of C: is '/' !! */
  if (home) dosdrive_add(3,home);
  dosdrive_set(2);  /* select C: for current drive */
}


int dosdrive_set ( int drive )
{
  if (drive>=CONFIG_MAX_DRIVES||!drives[drive].valid) return DNODRIVE;
  current_drive=drive;
  if (stddebug) fprintf(stddebug,"DOSDRIVE: set current drive to %c:\n",drive+'A');
  return 0;
}


int dosdrive_get ( void )
{
  if (stddebug) fprintf(stddebug,"DOSDRIVE: get current drive = %c:\n",current_drive+'A');
  return current_drive;
}


int dosdrive_getnumof ( void )
{
  if (stddebug) fprintf(stddebug,"DOSDRIVE: get number of drives = %d\n",dosdrives);
  return dosdrives;
}


int dosdrive_add ( int drive , char *root )
{
  if (drive>=CONFIG_MAX_DRIVES) return DNODRIVE;
  if (!realpath(root,path)) return DNOPATH;
  if (drives[drive].valid) {
    free(drives[drive].root);
    free(drives[drive].dir);
  }
  drives[drive].root=dosrun_strdup(path);
  drives[drive].valid=1;
  drives[drive].dir=dosrun_strdup("/");
  *(byte*)&drives[drive].acl=255;   /* ugly but fast :) set all ACLs for drive to enabled state. */
  if (drive+1>dosdrives) dosdrives=drive+1;
  return 0;
}


/* get pointer to drive's ACL structure */
struct drive_acl_struct *dosdrive_getacl ( int drive )
{
  if (drive>=CONFIG_MAX_DRIVES) return NULL;
  return &drives[drive].acl;
}


char *dosdrive_getroot ( int drive )
{
  if (drive) drive--; else drive=current_drive;
  if (drive>=CONFIG_MAX_DRIVES||!drives[drive].valid) return NULL;
  return drives[drive].root;
}
