/* INT21 implementation for DOSRUN
   (C)1999 LGB of DC, Gabor Lenart, lgb@vlug.vein.hu
   GNU/GPL code */

#include <stdio.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/vfs.h>
#include <time.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include "dosrun.h"
#include "memory.h"
#include "loader.h"
#include "filename.h"
#include "derrno.h"

#include "config_dosrun.h"

int dta_seg,dta_ofs;
int dos_version=(DOS_VER_MINOR<<8)+DOS_VER_MAJOR;
extern int exitcode;
extern word oldcs,oldip;

extern int int21_emu_execute ( void );

static byte temp_file_id=0;

#define ERRNO_TAB_SIZE 31
static const unsigned char errno_tab[ERRNO_TAB_SIZE] = {
     5, /* EPERM		 1	Operation not permitted */
     2, /* ENOENT		 2	No such file or directory */
     0, /* ESRCH		 3	No such process */
     0, /* EINTR		 4	Interrupted system call */
    23, /* EIO			 5	I/O error */
     0, /* ENXIO		 6	No such device or address */
     0, /* E2BIG		 7	Arg list too long */
     0, /* ENOEXEC	 8	Exec format error */
     6, /* EBADF		 9	Bad file number */
     0, /* ECHILD		10	No child processes */
     0, /* EAGAIN		11	Try again */
     8, /* ENOMEM		12	Out of memory */
     5, /* EACCES		13	Permission denied */
     0, /* EFAULT		14	Bad address */
     0, /* ENOTBLK	15	Block device required */
    32, /* EBUSY		16	Device or resource busy */
    80, /* EEXIST		17	File exists */
    17, /* EXDEV		18	Cross-device link */
     0, /* ENODEV		19	No such device */
     0, /* ENOTDIR	20	Not a directory */
     0, /* EISDIR		21	Is a directory */
     0, /* EINVAL		22	Invalid argument */
     4, /* ENFILE		23	File table overflow */
     4, /* EMFILE		24	Too many open files */
     0, /* ENOTTY		25	Not a typewriter */
    32, /* ETXTBSY	26	Text file busy */
     0, /* EFBIG		27	File too large */
     8, /* ENOSPC		28	No space left on device */
     0, /* ESPIPE		29	Illegal seek */
    19, /* EROFS		30	Read-only file system */
     5  /* EMLINK		31	Too many links */
};

int errno_u2d ( int err )
{
  if (err>ERRNO_TAB_SIZE) return 31;
  if (!errno_tab[err]) return 31;
  return errno_tab[err];
}

int check_avail_char ( int fd )
{
  struct timeval tv;
	int res;
  fd_set fds;
  tv.tv_sec=tv.tv_usec=0;
  FD_ZERO(&fds);
  FD_SET(fd,&fds);
  res=select(fd+1,&fds,0,0,&tv)?255:0;
	if (stddebug) fprintf(stddebug,"CHECK_CHAR_BUFFER: %d\n",res);
	return res;
}


#include "int21_misc.c"


int int21_emu ( void )
{
  switch (AH) {
    case 0x00 : if (stddebug) fprintf(stddebug,"Exit on int 21h AH=00h AL=??\n");
                return 0;
    case 0x4C : if (stddebug) fprintf(stddebug,"Exit on int 21h AH=4Ch AL=%02X\n",AL);
                exitcode=AL;
                return 0;
    case 0x31 : fprintf(stderr,"TSR programs are not supported by dosrun.\n");
                return 0;
    case 0x01 : read(0,&vm.regs.eax,1);
								putchar(AL);
								fflush(stdout);
								return 1;
    case 0x02 : putchar(DL);
                fflush(stdout);
								return 1;
    /* character direct I/O */
    case 0x06 : if (DL==255) {
                  if (check_avail_char(0)) {
										read(0,&vm.regs.eax,1);
										CLC;
									} else STC;
									return 1;
                }
								putchar(DL);
								fflush(stdout);
								return 1;
    /* char out without in */
    case 0x07 :
    case 0x08 : read(0,&vm.regs.eax,1);
                return 1;
    /* print '$' terminated string */
    case 0x09 : {
                  byte *p=DOS_PB(DS,DX);
                  while (*p!='$') putchar(*(p++));
                  fflush(stdout);
                  return 1;
                }
    /* get string */
    case 0x0A : {
                  byte *p=DOS_PB(DS,DX);
                  fgets(p+2,*p,stdin);
                  *(p+1)=strlen(p+2);
                  return 1;
                }
    case 0x0B : AL=check_avail_char(0);
                return 1;
    case 0x0C : if (AL!=1&&AL!=6&&AL!=7&&AL!=8&&AL!=10) return 1;
                AH=AL;
                (void)int21_emu();   /* ugly ? ;-) */
                AH=0x0C;
                return 1;
    case 0x0D : return 1;    /* reset block devices */
    case 0x0E : (void)dosdrive_set(DL);			
                AL=dosdrive_getnumof();
                return 1;
    case 0x0F : /* sily FCB functions, get lost ! */
    case 0x10 :
    case 0x11 :
    case 0x12 :
    case 0x13 :
    case 0x14 :
    case 0x15 :
    case 0x16 :
    case 0x17 :
    case 0x21 :
    case 0x22 :
    case 0x23 :
    case 0x24 :
    case 0x27 :
    case 0x28 :
    case 0x29 : AL=255;
                return 1;
    /* get current drive */
    case 0x19 : AL=dosdrive_get();
                return 1;
    /* set DTA */
    case 0x1A : dta_seg=DS;
                dta_ofs=DX;
                return 1;
    /* get disk drive information */
    case 0x1B : DL=0;  /* no break or return ! it's not a mistake */
    case 0x1C : if (DL) DL--; else DL=dosdrive_get();
                return 1;  /* TODO ! write the body of function */
    /* SET interrupt vector */
    case 0x25 : {
                  word *w=(word*)((unsigned int)AL<<2);
                  *w=DX;
                  *(w+1)=DS;
                  if (stddebug) fprintf(stddebug,"INT21: interrupt vector %02X chained to %04X:%04X\n",AL,DS,DX);
                  return 1;
		}
    /* get date */
    case 0x2A : {
                  time_t t=time(NULL);
                  struct tm *tim=localtime(&t);
                  AL=tim->tm_wday;
                  CX=tim->tm_year-80;
                  DH=tim->tm_mon+1;
                  DL=tim->tm_mday;
                  if (stddebug) fprintf(stddebug,"TIME: (int21) get date : %d-%d-%d (dow=%d)\n",CX+1980,DH,DL,AL);
                  return 1;
                }
    /* set date and time, Linux has got its own time mechanism do not allow to screw it by a DOS program */
    case 0x2D :
    case 0x2B : AL=0;
                return 1;
    /* get time */
    case 0x2C : {
                  time_t t=time(NULL);
                  struct tm *tim=localtime(&t);
                  struct timeval tv;
                  struct timezone tz;
                  gettimeofday(&tv,&tz);
                  CH=tim->tm_hour;
                  CL=tim->tm_min;
                  DH=tim->tm_sec;
                  DL=tv.tv_usec/10000;
                  if (stddebug) fprintf(stddebug,"TIME: (int21) get time : %02d:%02d:%02d.%02d\n",CH,CL,DH,DL);
                  return 1;
                }
    /* Set verify flag */
    case 0x2E : return 1; /* not output */
    /* get DTA address */
    case 0x2F : ES=dta_seg;
                BX=dta_ofs;
                return 1;
    /* Get DOS version */
    case 0x30 : AX=dos_version;
                return 1;
    /* get/set ctrl-break checking */
    case 0x33 : if (AL==0) DL=1;
                return 1;
    /* read interrupt vector */
    case 0x35 : {
                  word *w=(word*)((unsigned int)AL<<2);
                  BX=*w;
                  ES=*(w+1);
                  return 1;
                }
    /* Get free disk space */
    case 0x36 : {
                  char *p=dosdrive_getroot(DL);
                  struct statfs st;
                  if (!p||statfs(p,&st)) {
                    AX=0xFFFF;
                    return 1;
                  }
                  CX=st.f_bsize;  /* bytes per sector */
                  AX=st.f_blocks/65532+1;   /* sectors per cluster */
                  BX=st.f_bavail/AX;  /* free clusters */
                  DX=st.f_blocks/AX;  /* clusters */
                  return 1;
                }
    /* switch char ?! */
    case 0x37 : if (AL==0) { 
                  AL='-';
                  return 1;
		}
    /* get/set country information */
    case 0x38 : if (DX==65535) { /* set */ 
                  CLC;
                  return 1;
                } else { /* get */
                  byte *p=DOS_PB(DS,DX);
                  bzero(p,34);
                  *(word*)p=2; /* japan style dates : year-month-day */
                  *(p+2)='F';
                  *(p+3)='t'; /* hungarian forint */
                  *(p+7)=',';
                  *(p+9)='.';
                  *(p+11)='-';
                  *(p+13)=':';
                  *(p+15)=3;  /* X Ft */
                  *(p+16)=8;
                  *(p+17)=1;
                  *(p+22)=',';
                  *(word*)(p+18)=INT21_38_CAPITALPROC_OFS;  /* blah, seems vc.com uses it */
                  *(word*)(p+20)=INT21_38_CAPITALPROC_SEG;
                  return 1;
                }
    /* Make directory */
    case 0x39 : {
                  int r=path_d2u(DOS_PB(DS,DX),CREATE_ACCESS);
                  if (r) {
                    AX=DERRNO(r);
                    STC;
                    return 1;
                  }
                  if (mkdir(path,0777)) {
                    AX=DERRNO(errno_u2d(errno));
                    STC;
                  } else CLC;
                  return 1;
                }
    /* Delete directory */
    case 0x3A : {
                  int r=path_d2u(DOS_PB(DS,DX),DELETE_ACCESS);
                  if (r) {
                    AX=DERRNO(r);
                    STC;
                    return 1;
                  }
                  if (rmdir(path)) {
                    AX=DERRNO(errno_u2d(errno));
                    STC;
                  } else CLC;
                  return 1;
                  }
    /* Chdir */
    case 0x3B : {
                  int r=set_dosdir(DOS_PB(DS,DX));
                  if (r) {
                    AX=DERRNO(r);
                    STC;
                  } else CLC;
                  return 1;
                }
    /* Create file */
    case 0x3C : {
                  int fd;
                  int r=path_d2u(DOS_PB(DS,DX),CREATE_ACCESS);
                  if (r) {
                    AX=DERRNO(r);
                    STC;
                    return 1;
                  }
                  fd=open(path,O_RDWR|O_CREAT|O_TRUNC,CONFIG_CREATED_FILE_MOD);
                  if (fd==-1) {
                    if (stddebug) fprintf(stddebug,"FILEIO: cannot create file %s (%s).\n",path,strerror(errno));
                    AX=DERRNO(errno_u2d(errno));
                    STC;
                  } else {
                    AX=fd;
                    CLC;
                  }
                  return 1;
                }
    /* Open file */
    case 0x3D : {
                  int fd,mode,ac,r;
                  switch (vm.regs.eax&7) {
                    case 0  : mode=O_RDONLY; ac=READ_ACCESS;  break;
                    case 1  : mode=O_WRONLY; ac=WRITE_ACCESS; break;
                    default : mode=O_RDWR;   ac=READ_ACCESS|WRITE_ACCESS; break;
                  }
                  r=path_d2u(DOS_PB(DS,DX),ac);
                  if (r) {
                    AX=DERRNO(r);
                    STC;
                    return 1;
                  }
                  fd=open(path,mode);
                  if (fd==-1) {
                    if (stddebug) fprintf(stddebug,"FILEIO: cannot open file %s (%s).\n",path,strerror(errno));
                    AX=DERRNO(errno_u2d(errno));
                    STC;
                  } else {
                    AX=fd;
                    CLC;
                  }
                  return 1;
                 }
    /* Close file */
    case 0x3E : if (close(BX)) {
                  STC;
                  AX=DERRNO(errno_u2d(errno));
                } else CLC;
                return 1;
    /* Read from file */
    case 0x3F : {
                  int ret;
                  ret=read(BX,DOS_PB(DS,DX),CX);
                  if (ret==-1) {
                    STC;
                    AX=DERRNO(errno_u2d(errno));
                  } else {
                    CLC;
                    AX=ret;
                  }
                  return 1;
                }
    /* Write into file */
    case 0x40 : {
                  int ret;
                  ret=write(BX,DOS_PB(DS,DX),CX);
                  if (ret==-1) {
                    STC;
                    AX=DERRNO(errno_u2d(errno));
                  } else {
                    CLC;
                    AX=ret;
                  }
                  return 1;
                }
    /* Delete file */
    case 0x41 : {
                  int r=path_d2u(DOS_PB(DS,DX),DELETE_ACCESS);
                  if (r) {
                    AX=DERRNO(r);
                    STC;
                    return 1;
                  }
                  if (unlink(path)) {
                    STC;
                    AX=DERRNO(errno_u2d(errno));
                  } else CLC;
                  return 1;
                }
    /* Seek file */
    case 0x42 : {
                  int ret,mode;
                  switch (AL) {
                    case 0  : mode=SEEK_SET; break;
                    case 1  : mode=SEEK_CUR; break;
                    case 2  : mode=SEEK_END; break;
                    default : AX=DERRNO(DNOSYS);
                              STC;
                              return 1;
                  }
                  ret=lseek(BX,DOS_LONG(CX,DX),mode);
                  if (ret==-1) {
                    STC;
                    AX=DERRNO(errno_u2d(errno));
                  } else {
                    CLC;
                    DX=ret>>16;
                    AX=ret&0xffff;
                  }
                  return 1;
                }
    /* get/set file attributes */
    case 0x43 : switch (AL) {
                  case 0x00 : { /* get */
										struct stat st;
										int r=path_d2u(DOS_PB(DS,DX),STAT_ACCESS);
										if (r) {
											AX=DERRNO(r);
											STC;
											return 1;
										}
                    if (stat(path,&st)) {
                      AX=DERRNO(errno_u2d(errno));
                      STC;
                    } else {
                      int r=0;
                      if ((S_IWUSR&st.st_mode)==0) r|=1;
                      if (S_ISDIR(st.st_mode)) r|=16;
                      CX=r;
                      CLC;
                    }
                    return 1;
                  }	
                  case 0x01 : { /* set */
										struct stat st;
                    int r=path_d2u(DOS_PB(DS,DX),CHMOD_ACCESS);
                    if (r) {
                      AX=DERRNO(r);
                      STC;
                      return 1;
                    }
                    if (stat(path,&st)) {
                      AX=DERRNO(errno_u2d(errno));
                      STC;
                      return 1;
                    }
                    if (S_ISDIR(st.st_mode)) {
                      AX=DERRNO(DNOACCESS);
                      STC;
                      return 1;
                    }
                    if (((S_IWUSR&st.st_mode)&&(CX&1))||((!S_IWUSR&st.st_mode)&&(!CX&1))) {
                      if (st.st_mode&S_IWUSR) st.st_mode&=~S_IWUSR; else st.st_mode|=S_IWUSR;
                      if (chmod(path,st.st_mode)) {
                        AX=DERRNO(errno_u2d(errno));
                        STC;
                        return 1;
                      }
                    }
                    CLC;   /* fake : other attr changing was ok */
                    return 1;
									}
                }
                STC;          /* invalid function */
                AX=DERRNO(DNOSYS);
                return 1;
    /* IOCTL functions */
    case 0x44 : return int21_emu_ioctl();
    /* duplicate file handle ?? */
    case 0x45 : {
                  int r=dup(BX);
                  if (r==-1) {
                    STC;
                    AX=DERRNO(errno_u2d(errno));
                  } else {
                    CLC;
                    AX=r;
                  }
                  return 1;
                }
    /* dup2()  ?? */
    case 0x46 : if (dup2(BX,CX)) {
                  STC;
                  AX=DERRNO(errno_u2d(errno));
                } else CLC;
                return 1;
    /* get current directory */
    case 0x47 : {
                  int r=get_dosdir(DL,DOS_PB(DS,SI));
                  if (r) {
                    STC;
                    AX=DERRNO(r);
                  } else CLC;
                  return 1;
                }
    /* allocate RAM */
    case 0x48 : {
                  int r=dosmem_alloc(BX,SEARCH_BEST,pid);
                  if (r<=0) {
                    STC;
                    AX=DERRNO(-r);
                    BX=dosmem_largest();
                  } else {
                    CLC;
                    AX=r>>4;
                  }
                  return 1;
                }
    /* free RAM */
    case 0x49 : {
                  int r=dosmem_free(ES<<4,pid);
                  if (r) {
                    STC;
                    AX=DERRNO(r);
                  } else CLC;
                  return 1;
                }
    /* realloc RAM */
    case 0x4A : {
                  int r=dosmem_realloc(ES<<4,BX,pid);
                  if (r) {
                    STC;
                    AX=DERRNO(r);
                    BX=dosmem_largest();
                  } else CLC;
                  return 1;
                }
    /* execute/load program */
    case 0x4B : return int21_emu_execute();
    /* find first directory entry */
    case 0x4E : return int21_emu_findfirst();
    /* find next directory entry */
    case 0x4F : return int21_emu_findnext();
    /* set PID (PSP address) */
    case 0x50 : return 1;  /* FAKE, return ! */
    /* rename/move file */
    case 0x56 : {
                  byte opath[DOS_PATH_MAXSIZE];
                  int r;
                  r=path_d2u(DOS_PB(DS,DX),MOVE_ACCESS);   /* old name */
                  if (r) {
                    STC;
                    AX=DERRNO(r);
                    return 1;
                  }
                  strcpy(opath,path);  /* opath <- path */
                  r=path_d2u(DOS_PB(ES,DI),MOVE_ACCESS);   /* new name */
                  if (r) {
                    STC;
                    AX=DERRNO(r);
                    return 1;
                  }
                  if (rename(opath,path)) {
                    STC;
                    AX=DERRNO(errno_u2d(errno));
                  } CLC;
                  return 1;
                }
    /* get or set (?) last modification time of a file */
    case 0x57 : {
                  if (AL==0) { /* get */
                    struct stat st;
                    struct tm *tim;
                    if (fstat(BX,&st)) {
                      AX=DERRNO(errno_u2d(errno));
                      STC;
                      return 1;
                    }
                    tim=localtime(&st.st_mtime);
                    if (stddebug) fprintf(stddebug,"FILEIO: file fd %d modificated last on %d-%d-%d at %d:%d:%d.\n",BX,tim->tm_year,tim->tm_mon,tim->tm_mday,tim->tm_hour,tim->tm_min,tim->tm_sec);
                    CX=(tim->tm_sec>>1)+(tim->tm_min<<5)+(tim->tm_hour<<11);
                    DX=tim->tm_mday+((tim->tm_mon+1)<<5)+((tim->tm_year-80)<<9);
                    CLC;
                    return 1;
                  }
                  AX=DERRNO(DNOACCESS);  /* setting is not allowed */
                  STC;
                  return 1;
                }
    /* memory allocation strategy */
    case 0x58 : if (AL==0) {  /* read: best fit */
                  AX=1;
                  CLC;
                  return 1;
                } else {      /* write: NOT allowed now ! */
                  STC;
                  AX=DERRNO(DNOSYS);
                  return 1;
                }
    /* get extended error code */
		case 0x59 : if (stddebug) fprintf(stddebug,"59: derrno=%d\n",derrno);
								AX=derrno;
                BH=13; /* other error */
								BL=6; /* skip error */
                CH=1; /* unknown */
                return 1;
    /* create temporary file */
    case 0x5A : {
                  int fd,r;
                  char buffer[14];
                  char *p=DOS_PB(DS,DX);
                  sprintf(buffer,"__temp__.%d",temp_file_id++);
                  if (!*p||p[strlen(p)-1]!='\\') strcat(p,"\\");
                  strcat(p,buffer);
                  r=path_d2u(p,CREATE_ACCESS);
                  if (r) {
                    AX=DERRNO(r);
                    STC;
                    return 1;
                  }
                  fd=open(path,O_RDWR|O_CREAT|O_TRUNC,CONFIG_CREATED_FILE_MOD);
                  if (fd==-1) {
                    if (stddebug) fprintf(stddebug,"FILEIO: cannot create temporaty file %s (%s).\n",path,strerror(errno));
                    AX=DERRNO(errno_u2d(errno));
                    STC;
                  } else {
                    AX=fd;
                    CLC;
                  }
                  return 1;
                }
    /* get current PSP address (PID) */
    case 0x51 :
    case 0x62 : BX=pid;
                return 1;
    /* get double byte character */
    case 0x63 : if (AL==0) AL=0xFF;
                STC;
                return 1;
    /* country information and character captailization */
    case 0x65 : if (AL<8) {  /* pass dummy country info */
                  BX=-1;
                  DX=-1;
                  STC;
                  return 1;
                }
                break;
	  /* extended open/create, fast and ugly implementation :( */
		case 0x6C : {
									int ac,mode;
									switch (AL) {
									  case 0 : mode=O_RDONLY;
														 ac=READ_ACCESS;
														 break;
										case 1 : mode=O_WRONLY;
														 ac=WRITE_ACCESS;
														 break;
										default: mode=O_RDWR;
														 ac=WRITE_ACCESS;
														 break;
									}
                  if (!DL&15) mode|=O_EXCL;
									if (DL&18) {
									  mode|=O_TRUNC|O_CREAT;
										ac|=CREATE_ACCESS;
									}
								  ac=path_d2u(DOS_PB(DS,SI),ac);	
									if (ac) {
									  AX=DERRNO(ac);
										STC;
										return 1;
									}
									ac=open(path,mode,CONFIG_CREATED_FILE_MOD);
									if (ac==-1) {
									  AX=DERRNO(errno_u2d(errno));
										STC;
										return 1;
									}
									AX=ac;
									CLC;
								}
  } /* switch */
#ifdef CONFIG_DEBUG_UNIMPLEMENTED_FUNC
  if (stddebug) fprintf(stddebug,"INT21: unimplemented function (CS:IP=%04X:%04X) AX=%04X BX=%04X ES=%04X DX=%04X SI=%04X\n",oldcs,oldip,AX,BX,ES,DX,SI);
#endif
  return 1;
}

