/*
 *                            COM/EXE loader for DOSRUN
 *                  (C)1999 LGB (Gabor Lenart), lgb@vlug.vein.hu
 *      EXE loader is heavily based on Arpi/ESP-Team's code from old dosrun
 *                              GNU/GPL source code.
 */


#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include "dosrun.h"
#include "memory.h"
#include "derrno.h"
#include "loader.h"
#include "config_dosrun.h"

extern int dta_seg,dta_ofs;
int pid;
static int env_length;
static word *env_ext;


/* 1. As far as I know DOS allocates the WHOLE memory for a program, and
      applications must resize its memory. So I use SEARCH_NOSPLIT to allocate
      the whole block for us. Also known by me that DOS allocates the last
      block, so I use SEARCH_LAST parameter. End of this block is written
      into PSP+2 too. - LGB
*/

static int load_exe ( int fd )
{
  word header[14];
  word temp;
  int exesize,headsize,addr,base,minmem;
  read(fd,&header,sizeof(header));
  if (header[1]) exesize=((header[2]-1)<<9)+header[1]; else
    exesize=header[2]<<9;
  headsize=header[4]<<4;
  minmem=header[5]<<4;
  addr=dosmem_alloc((exesize-headsize+0x100+64)>>4,SEARCH_LAST|SEARCH_NOSPLIT,8);
  if (!addr) return -DNOMEM;
  if (stddebug) fprintf(stddebug,"Loading EXE file to %05Xh.\n",addr);
  lseek(fd,headsize,SEEK_SET);
  base=(addr+0x100)>>4;
  if (read(fd,(void*)(addr+0x100),exesize-headsize)!=exesize-headsize) return -DBADFORMAT;
  if (header[3]) { /* some relocation code */
    word *reloc=(word*)malloc(header[3]<<2);      /* malloc, not dosrun_malloc */
    word *r=reloc;
    int a;
    if (!reloc) return -DNOMEM;
    if (stddebug) fprintf(stddebug,"Loading %d relocation points.\n",header[3]);
    lseek(fd,header[12],SEEK_SET);
    read(fd,reloc,header[3]<<2);
    for (a=0;a<header[3];a++) {
      word *rp=(word*)(((base+r[1])<<4)+r[0]);
      if ((int)rp>=CONFIG_DOSMEM_SIZE<<9) return -DBADFORMAT;
      *rp+=base;
      r+=2;
    } /* for */
    free(reloc);
  } /* if */
  DS=ES=base-0x10; /* DS and ES points to PSP by default */
  EIP=header[10];
  temp=header[11]+base;
  CS=temp;
  ESP=header[8];
  temp=header[7]+base;
  SS=temp;
  if (stddebug) fprintf(stddebug,"Loading EXE is ready.\n");
  return addr;
}


static int load_com ( int fd )
{
  int addr=dosmem_alloc(0x1000,SEARCH_LAST|SEARCH_NOSPLIT,8);
  if (!addr) return -DNOMEM;
  if (stddebug) fprintf(stddebug,"Loading COM file to %05Xh.\n",addr);
  read(fd,(void*)(addr+0x100),0xFF00);
  ES=DS=SS=CS=addr>>4;
  ESP=0xFFFE;
  EIP=0x100;
  *DOS_PW(SS,0xFFFE)=0; /* push 0 */
  if (stddebug) fprintf(stddebug,"Loading COM is ready.\n");
  return addr;
}


int load_prg ( char *name , char *env )
{
  int fd=open(name,O_RDONLY),ret;
  word signature;
  int dosenv;
  if (fd==-1) return -DNOENT;
  read(fd,&signature,2);
  lseek(fd,0,SEEK_SET);
  if (env) {
    dosenv=dosmem_alloc((env_length>>4)+1,SEARCH_BEST,8);
    memcpy((byte*)dosenv,env,env_length);
  }
  if (signature==0x4D5A||signature==0x5A4D) ret=load_exe(fd); else ret=load_com(fd);
  close(fd);
  if (ret<=0) return ret;
  pid=ret>>4;  /* DOS use PSP segment address as PID, AFAIK */
  dosmem_setowner(ret,pid);  /* reowns memory area */
  bzero((byte*)ret,0x100); /* fill PSP with zero, thx NFL */
  *(word*)ret=0x20CD;  /* int 20h */
  *(word*)(ret+2)=(ret>>4)+MCB_X_SIZE(ret-MCB_SIZE);  /* memtop for appilcation */
  if (env) {
    dosmem_setowner(dosenv,pid);
    *(word*)(ret+0x2C)=dosenv>>4;  /* set up envirnoment pointer in PSP */
  }
  dta_seg=ret>>4;   /* setup default DTA */
  dta_ofs=0x80;
  if (stddebug) fprintf(stddebug,"CS:IP=%04X:%04X SS:SP=%04X:%04X PSP=%04X\n",CS,IP,SS,SP,ret>>4);
  return ret;
}



static __inline__ char *append_env ( char *p , char *str , byte tail )
{
  while (*str) *(p++)=*(str++);
  *(p++)=tail;
  return p;
}



/* Builds envirnoment block for the FIRST run.
   Pass NULL at the first call. Passing NULL as val represents extended
   envirnoment string (program name in envirnoment block) use this ONLY
   after you passed ALL 'normal' envirnoment strings ! */
char *env_pass_string ( char *env , char *str , char *val )
{
  int size=strlen(str)+1,firstext=0;
  char *p;
  if (!env) {
    env_length=0;
    env_ext=NULL;
  }
  if (val) size+=strlen(val)+1;
  if (!env_ext&&!val) {
    firstext=1;
    size+=3;
  }
  env=dosrun_realloc(env,env_length+size);
  p=env+env_length;
  env_length+=size;
  if (firstext) {
    *(p++)='\0';
    env_ext=(word*)p;
    p+=2;
    *env_ext=0;
  }
  if (val) {
    p=append_env(p,str,'=');
    p=append_env(p,val,0);
  } else {
    (*env_ext)++;
    p=append_env(p,str,0);
  }
  return env;
}



/* for testing env_pass_string function ...  */
void env_dump ( char *env )
{
  if (!stddebug) return;
  while (*env) {
    printf("NORM_STRING: ");
    while (*env) fprintf(stddebug,"%c",*(env++));
    env++;
    fprintf(stddebug,"\n");
  }
  if (env_ext) {
    word num;
    env++;
    num=*((word*)env);
    fprintf(stddebug,"%d ext strings.\n",num);
    env+=2;
    while (num--) {
      fprintf(stddebug,"EXT_STRING: ");
      while (*env) fprintf(stddebug,"%c",*(env++));
      env++;
      fprintf(stddebug,"\n");
    }
  }
  fprintf(stddebug,"Total size = %d bytes.\n",env_length);
}


void pass_commandline ( int addr , int argc , char ** argv )
{
  byte *p=(byte*)(addr+0x81);
  byte *p0=(byte*)(addr+0x80);
  int a;
  *p='\0';
  for (a=0;a<argc;a++) {
    byte *q=argv[a];
    while (*q) *(p++)=*(q++);
    if (a+1!=argc) *(p++)=' ';
  }
  *p0=(byte)(p-p0-1);
  *(p++)=13;
  *p='\0';  /* just for us */
  if (stddebug) fprintf(stddebug,"ARG SIZE=%d ARG=\"%s\"\n",*p0,p0+1);
}


int int21_emu_execute ( void )
{
  return 1;
}
