/*
 *   Command line and configuration file parser for DOSRUN. GNU/GPL source.
 *                (C)1999 LGB, Gabor Lenart, lgb@vlug.vein.hu
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include "config_dosrun.h"
#include "dosrun.h"
#include "options.h"
#include "loader.h"
#include "xms.h"
#include "filename.h"

#define MAX_PAR_PER_OPT 32

struct config_struct config;

extern char *env;

struct options_struct {
  char *s;
  int  min,max;
  int (*func)(int,int,char**);    /* num of params, on/off switch, p1, p2, p3, p4 */
  char *comment;
};

#include "options_options.c"

static const struct options_struct options[] = {
  { "env"    , 1 , MAX_PAR_PER_OPT, optf_env     , "pass an environment string to DOS application"  },
#ifdef CONFIG_FULL_SCREEN
  { "fullscr", 0 , 0, optf_fullscr , "full screen mode (*)"                               },
#ifdef CONFIG_AUTO_FULL_SCREEN  
  { "afullscr",0 , 0, optf_afullscr,"switching into fullscreen mode automatically (*)" },
#endif
#endif
  { "debug"  , 1 , 1, optf_debug   , "debugging information on stderr"                },
  { "dosver" , 1 , 1, optf_dosver  , "set emulated dos version"                       },
  { "drive"  , 2 , 3, optf_drive   , "assign drive letter to a directory structure"   },
  { "version", 0 , 0, optf_version , "print version information about dosrun"         },
#ifdef CONFIG_XMS
  { "xms"    , 3 , 3, optf_xms     , "tune parameters for XMS emulation (*)"          },
#endif
  {NULL,0,0,NULL}
};


/* return value of this function is the first number of argv which is NOT option but the program name user wants to run */
int options_parser ( int argc , char ** argv )
{
  int r=1;
  while (r<argc&&*argv[r]=='-'&&strcmp(argv[r],"--")) {
    const struct options_struct *opt=options;
    char *s=argv[r]+1;
    int enable=1;
    if (*s=='n'&&s[1]=='o') enable=0,s+=2;
    while (opt->s&&strcmp(opt->s,s)) opt++;
    if (!opt->s) {
      fprintf(stderr,"Unknown option : %s\n",argv[r]);
      return -1;
    } else {
      char *p[MAX_PAR_PER_OPT];
      int n=0;
//      bzero(p,sizeof(p));
      r++;
      if (opt->max&&enable)
        while (r<argc&&n<opt->max&&*argv[r]!='-'&&strcmp(argv[r],"--"))
	  p[n++]=argv[r++];
      if (enable&&n<opt->min) {
	fprintf(stderr,"Option -%s requires at least %d parameters (or unexepcted end of command line).\n",opt->s,opt->min);
	return -1;
      }     
      if ((opt->func)(n,enable,p)) return -1;
    }
  }   /* while */
  if (r<argc&&!strcmp(argv[r],"--")) r++;
  if (r>=argc) {
    fprintf(stderr,"No program name specified to run.\n");
    return -1;
  }
  if (stddebug) fprintf(stddebug,"Program name to run = %s\n",argv[r]);
  return r;
}




void options_help ( char *prg )
{
  int a=0;
  fprintf(stderr,"DOSRUN v%s (C)1999 Arpi/ESP-team, LGB/DC, NFL
Free software under terms written in GPL2, NO WARRANTY !\n
Usage : %s [options] [--] <program-name> [program parameters]\n
Options : (options marked with '*' can be used with 'no' modifier, eg. -noxms)\n\n",DOSRUN_VERSION,prg);
  while (options[a].s) {
    char *p=options[a].s,b=12;
    printf("  -");
    while (b--)
      if (*p) putchar(*(p++)); else putchar(' ');
    printf("%s\n",options[a].comment);
    a++;
  }
  putchar('\n');
}
